// -*- C++ -*-
#ifndef HERWIG_HwDecayHandler_H
#define HERWIG_HwDecayHandler_H
//
// This is the declaration of the HwDecayHandler class.
//
#include "ThePEG/Handlers/DecayHandler.h"
#include "Herwig++/Hadronization/PartonicHadronizer.h"

using namespace ThePEG;
namespace Herwig {

  /** \ingroup Decay
   * The <code>HwDecayHandler</code> is the Herwig++ decay handler which 
   *  administers the decays of unstable particles in Herwig++. It
   * is derived from ThePEG::DecayHandler and includes a different handle
   * method in order to simulate decays including spin correlations.
   *
   * The handle method decays all particles in the current step, including
   * spin correlations. Another feature of the DecayHandler is that it correctly
   * handles mutlistep decays where a Decayer supplys intermediate decay products
   * in addition to the outgoing particles.
   *
   * @see ThePEG::StepHandler
   * @see ThePEG::CollisionHandler
   * @see ThePEG::SubProcessHandler
   * @see ThePEG::DecayHandler
   * 
   */

class HwDecayHandler: public DecayHandler {

public:

  /** @name Standard constructors and destructors. */
  //@{
  /**
   * Default constructor
   */
  inline HwDecayHandler();

  /**
   * Copy constructor
   */
  inline HwDecayHandler(const HwDecayHandler &x);

  /**
   * Destructor
   */
  virtual ~HwDecayHandler();
  //@}

public:

  /**
   * Look through all \a tagged particled and decay all unstable ones.
   * @param ch the PartialCollisionHandler (now EventHandler) in charge of the generation.
   * @param tagged the vector of particles to consider. If empty, all
   * final state particles in the current Step is considered.
   * @param hint a possible Hint which is ignored in this implementation.
   */
  virtual void handle(EventHandler & ch, const tPVector & tagged,
		      const Hint & hint)
    ThePEG_THROW_SPEC((Veto, Stop, Exception));

  /**
   * Perform the decay of one unstable particle.
   * @param parent the particle to be decayed.
   * @param s the Step where decay products are inserted.
   */
  virtual void performDecay(tPPtr parent, Step & s) const
    ThePEG_THROW_SPEC((Veto, Exception));
  
  /**
   * add the decay products of in intermediate particle produced in a decay
   * @param parent the particle which has been decayed.
   * @param s the Step where decay products are inserted.
   */
  void addDecayedParticle(tPPtr parent, Step & s) const
    ThePEG_THROW_SPEC((Veto, Exception));

  /**
   * Standard Init function
   */
  static void Init();

public:

  /** @name Functions used by the persistent I/O system. */
  //@{
  /**
   * Function used to write out object persistently.
   * @param os the persistent output stream written to.
   */
  void persistentOutput(PersistentOStream & os) const;

  /**
   * Function used to read in object persistently.
   * @param is the persistent input stream read from.
   * @param version the version number of the object when written.
   */
  void persistentInput(PersistentIStream & is, int version);
  //@}

protected:

  /** @name Clone Methods. */
  //@{
  /**
   * Make a simple clone of this object.
   * @return a pointer to the new object.
   */
  inline virtual IBPtr clone() const;

  /** Make a clone of this object, possibly modifying the cloned object
   * to make it sane.
   * @return a pointer to the new object.
   */
  inline virtual IBPtr fullclone() const;
  //@}

  /** @name Standard Interfaced functions. */
  //@{
  /**
   * Check sanity of the object during the setup phase.
   */
  inline virtual void doupdate() throw(UpdateException);

  /**
   * Initialize this object after the setup phase before saving and
   * EventGenerator to disk.
   * @throws InitException if object could not be initialized properly.
   */
  inline virtual void doinit() throw(InitException);

  /**
   * Finalize this object. Called in the run phase just after a
   * run has ended. Used eg. to write out statistics.
   */
  inline virtual void dofinish();

  /**
   * Rebind pointer to other Interfaced objects. Called in the setup phase
   * after all objects used in an EventGenerator has been cloned so that
   * the pointers will refer to the cloned objects afterwards.
   * @param trans a TranslationMap relating the original objects to
   * their respective clones.
   * @throws RebindException if no cloned object was found for a given pointer.
   */
  inline virtual void rebind(const TranslationMap & trans)
    throw(RebindException);

  /**
   * Return a vector of all pointers to Interfaced objects used in
   * this object.
   * @return a vector of pointers.
   */
  inline virtual IVector getReferences();
  //@}


private:

  /**
   * Describe a concrete class with persistent date/
   */
  static ClassDescription<HwDecayHandler> initHwDecayHandler;

  /**
   *  Private and non-existent assignment operator.
   */
  HwDecayHandler & operator=(const HwDecayHandler &);

private:

  /**
   * Pointer to the object for the hadronization of partonic decays
   */
  PartonicHadronizerPtr _partonhad;

  /**
   *  Number of tries to regenerate a partonic decay to sucessfully hadronize it
   */
  unsigned int _hadtry;
};
}

namespace ThePEG {

/**
 * This template specialization informs ThePEG about the base class of
 * Hw64Decayer.
 */
template <>
struct BaseClassTrait<Herwig::HwDecayHandler,1> {
  /** Typedef of the base class of Hw64Decayer. */
  typedef DecayHandler NthBase;
};

/**
 * This template specialization informs ThePEG about the name of the
 * Hw64Decayer class.
 */
template <>
struct ClassTraits<Herwig::HwDecayHandler>: public ClassTraitsBase<Herwig::HwDecayHandler> {
  /** Return the class name. */
  static string className() { return "Herwig++::HwDecayHandler"; }
  /** Return the name of the shared library to be loaded to get
   * access to this class and every other class it uses
   * (except the base class).
   */
  static string library() { return ""; }
};

}

#include "HwDecayHandler.icc"
#ifndef ThePEG_TEMPLATES_IN_CC_FILE
// #include "HwDecayHandler.tcc"
#endif

#endif /* HERWIG_HwDecayHandler_H */
