// -*- C++ -*-
#ifndef HERWIG_PartonSplitter_H
#define HERWIG_PartonSplitter_H

#include "CluHadConfig.h"
#include <ThePEG/Handlers/HandlerBase.h>
#include "Herwig++/Utilities/GlobalParameters.h"


namespace Herwig {


using namespace ThePEG;


/** \ingroup Hadronization
 *  \class PartonSplitter
 *  \brief This class splits the gluons from the end of the shower.
 *  \author Philip Stephens
 *  \author Alberto Ribon
 * 
 *  This class does all of the nonperturbative parton splittings needed 
 *  immediately after the end of the showering (both initial and final),
 *  as very first step of the cluster hadronization.
 *
 *  See also:
 *  GlobalParameters.h.
 */
class PartonSplitter: public ThePEG::HandlerBase {

public:

  /** @name Standard constructors and destructors. */
  //@{
  /**
   * The default constructor.
   */
  inline PartonSplitter();

  /**
   * The copy constructor.
   */
  inline PartonSplitter(const PartonSplitter &);

  /**
   * The destructor.
   */
  virtual ~PartonSplitter();
  //@}

public:

  /**
   * This method does the nonperturbative splitting of:
   * time-like gluons. At the end of the shower the gluons should be
   * on a "physical" mass shell and should therefore be time-like.
   * @param tagged The tagged particles to be split
   * @param pstep Pointer to the step
   * @return The particles which were not split and the products of splitting.
   */
  tPVector split(const tPVector & tagged, tStepPtr pstep);
 
public:

  /** @name Functions used by the persistent I/O system. */
  //@{
  /**
   * Function used to write out object persistently.
   * @param os the persistent output stream written to.
   */
  void persistentOutput(PersistentOStream & os) const;

  /**
   * Function used to read in object persistently.
   * @param is the persistent input stream read from.
   * @param version the version number of the object when written.
   */
  void persistentInput(PersistentIStream & is, int version);
  //@}

  /**
   * The standard Init function used to initialize the interfaces.
   * Called exactly once for each class by the class description system
   * before the main function starts or
   * when this class is dynamically loaded.
   */
  static void Init();

protected:

  /** @name Clone Methods. */
  //@{
  /**
   * Make a simple clone of this object.
   * @return a pointer to the new object.
   */
  inline virtual IBPtr clone() const;

  /** Make a clone of this object, possibly modifying the cloned object
   * to make it sane.
   * @return a pointer to the new object.
   */
  inline virtual IBPtr fullclone() const;
  //@}

protected:

  /** @name Standard Interfaced functions. */
  //@{
  /**
   * Check sanity of the object during the setup phase.
   */
  inline virtual void doupdate() throw(UpdateException);

  /**
   * Initialize this object after the setup phase before saving an
   * EventGenerator to disk.
   * @throws InitException if object could not be initialized properly.
   */
  inline virtual void doinit() throw(InitException);

  /**
   * Finalize this object. Called in the run phase just after a
   * run has ended. Used eg. to write out statistics.
   */
  inline virtual void dofinish();

  /**
   * Rebind pointer to other Interfaced objects. Called in the setup phase
   * after all objects used in an EventGenerator has been cloned so that
   * the pointers will refer to the cloned objects afterwards.
   * @param trans a TranslationMap relating the original objects to
   * their respective clones.
   * @throws RebindException if no cloned object was found for a given
   * pointer.
   */
  inline virtual void rebind(const TranslationMap & trans)
    throw(RebindException);

  /**
   * Return a vector of all pointers to Interfaced objects used in this
   * object.
   * @return a vector of pointers.
   */
  inline virtual IVector getReferences();
  //@}

private:

  /**
   * Describe a concrete class with persistent data.
   */
  static ClassDescription<PartonSplitter> initPartonSplitter;

  /**
   * Private and non-existent assignment operator.
   */
  PartonSplitter & operator=(const PartonSplitter &);

  /**
   * Given in input a pointer to a time-like gluon, it forces
   * a nonperturbative quark - anti-quark splitting, returning
   * the pointers to these produced two new particles. 
   * If something wrong happens, it will returns null pointers.
   */
  void splitTimeLikeGluon(tcPPtr ptrGluon,                   // input        
			  PPtr & ptrQ, PPtr & ptrQbar);      // output
  
  /**
   * Given in input a pointer to a space-like gluon, it forces
   * a nonperturbative quark - anti-quark splitting, returning
   * the pointers to these produced two new particles. 
   * If something wrong happens, it will returns null pointers.
   */
  //void splitSpaceLikeGluon(tcPPtr ptrGluon,                  // input       
  //         		     PPtr & ptrQ, PPtr & ptrQbar);     // output
  
  /**
   * Given in input a pointer to a space-like sea quark, it forces 
   * a nonperturbative soft gluon emission, returning the pointers 
   * to the emitted gluon and the sea quark after the emission. 
   * If something wrong happens, it will return null pointers.
   */
  //void splitSpaceLikeSeaQuark(tcPPtr ptrSeaQ0,                   // input
  //			        PPtr & ptrGluon, PPtr & ptrSeaQ1); // output

  /**
   * Print full information for debugging.
   */
  void debuggingInfo(const tPVector & tagged, const set<tPPtr> & newPartons);

  /**
   * Pointer to a Herwig::GlobalParameters object for using global variables.
   */
  GlobParamPtr _globalParameters;  

};


}

namespace ThePEG {


/**
 * The following template specialization informs ThePEG about the
 * base class of PartonSplitter.
 */
template <>
struct BaseClassTrait<Herwig::PartonSplitter,1> {
  /** Typedef of the first base class of PartonSplitter. */
  typedef ThePEG::HandlerBase NthBase;
};

/**
 * The following template specialization informs ThePEG about the
 * name of this class and the shared object where it is defined.
 */
template <>
struct ClassTraits<Herwig::PartonSplitter>:
    public ClassTraitsBase<Herwig::PartonSplitter> {
  /** Return the class name.*/
  static string className() { return "Herwig++::PartonSplitter"; }
  /**
   * Return the name of the shared library to be loaded to get
   * access to this class and every other class it uses
   * (except the base class).
   */
  static string library() { return "libHwHadronization.so"; }
};

}

#include "PartonSplitter.icc"

#endif /* HERWIG_PartonSplitter_H */
