#ifndef HERWIG_UA5_H_
#define HERWIG_UA5_H_

#include <ThePEG/Handlers/MultipleInteractionHandler.h>
#include "Herwig++/Hadronization/ClusterFissioner.h"
#include "Herwig++/Hadronization/ClusterDecayer.h"
#include "Herwig++/Hadronization/PartonSplitter.h"
#include "Herwig++/Utilities/GlobalParameters.h"
#include <ThePEG/Handlers/EventHandler.h>
 

namespace Herwig {

using namespace ThePEG;

/** \ingroup UnderlyingEvent
 *
 *  This is the class definition for the UA5Handler. This 
 *  class is designed to generate an underlying event     
 *  based on the UA5 model. This is intended as a basic   
 *  underlying event model which will be superceded by a  
 *  new model in Herwig++.                                
 *                                                        
 *  This class interfaces with the Hadron decays and with 
 *  the cluster hadronization. To that end there is an    
 *  interface set up with the ClusterFissioner class and  
 *  with the ClusterDecayer class. The simple steps can   
 *  accessed through the following interfaces:            
 *                                                        
 *  pair<PPtr,PPtr>                                       
 *       ClusterDecayer::decayIntoTwoHadrons(tClusterPtr&)
 *  pair<pair<PPtr,PPtr>,pair<PPtr,PPtr> >                
 *       ClusterFissioner::cut(tClusterPtr&)              
 *  pair<PPtr,PPtr>                                       
 *       ClusterFissioner::produceCluster(...)            
 *                                                        
 *  The first method is used to split a cluster into two  
 *  hadrons. The returned value are the pointers to the   
 *  two new hadrons. The second method is used to split a 
 *  cluster C->C1+C2 or C->C1+H2 or C->H1+C2 or C->H1+H2  
 *  but the results depend on the mass of the input and   
 *  the mass generated by the distributions. If the given 
 *  distributions aren't sufficient, the third method can 
 *  be used to subvert the process. This will generate one
 *  cluster based on the input values. The second method  
 *  returns two pairs of particles. In each pair the first
 *  particle is one of the decay products, the second is  
 *  the particle drawn from the vacuum in order to create 
 *  the new particle. The produceCluster method returns a 
 *  similar things, the first is the new cluster and the  
 *  second is the flavour drawn from the vacuum.          
 */

class UA5Handler : public MultipleInteractionHandler {

public:

  /** @name Standard constructors and destructors. */
  //@{
  /**
   * The default constructor.
   */
   UA5Handler();

  /**
   * The copy constructor.
   */
   UA5Handler(const UA5Handler &);

  /**
   * The destructor.
   */
   virtual ~UA5Handler();
   //@}

  /**
   * The standard Init function used to initialize the interfaces.
   * Called exactly once for each class by the class description system
   * before the main function starts or
   * when this class is dynamically loaded.
   */
  static void Init();

public:

  /** @name Functions used by the persistent I/O system. */
  //@{
  /**
   * Function used to write out object persistently.
   * @param os the persistent output stream written to.
   */
  void persistentOutput(PersistentOStream & os) const;

  /**
   * Function used to read in object persistently.
   * @param is the persistent input stream read from.
   * @param version the version number of the object when written.
   */
  void persistentInput(PersistentIStream & is, int version);
  //@}

public:

  /**
   * This is the routine that starts the algorithm.
   */
  virtual void handle(EventHandler &ch, const tPVector &tagged,
		      const Hint &hint) 
    throw(Veto,Stop,Exception);

protected:

  /** @name Clone Methods. */
  //@{
  /**
   * Make a simple clone of this object.
   * @return a pointer to the new object.
   */
  inline virtual IBPtr clone() const;

  /** Make a clone of this object, possibly modifying the cloned object
   * to make it sane.
   * @return a pointer to the new object.
   */
  inline virtual IBPtr fullclone() const;
  //@}

protected:

  /** @name Standard Interfaced functions. */
  //@{
  /**
   * Check sanity of the object during the setup phase.
   */
  inline virtual void doupdate() throw(UpdateException);

  /**
   * Initialize this object after the setup phase before saving an
   * EventGenerator to disk.
   * @throws InitException if object could not be initialized properly.
   */
  inline virtual void doinit() throw(InitException);

  /**
   * Finalize this object. Called in the run phase just after a
   * run has ended. Used eg. to write out statistics.
   */
  inline virtual void dofinish();

  /**
   * Rebind pointer to other Interfaced objects. Called in the setup phase
   * after all objects used in an EventGenerator has been cloned so that
   * the pointers will refer to the cloned objects afterwards.
   * @param trans a TranslationMap relating the original objects to
   * their respective clones.
   * @throws RebindException if no cloned object was found for a given
   * pointer.
   */
  inline virtual void rebind(const TranslationMap & trans)
    throw(RebindException);

  /**
   * Return a vector of all pointers to Interfaced objects used in this
   * object.
   * @return a vector of pointers.
   */
  inline virtual IVector getReferences();
  //@}

private:

   ParticleVector decayHadron(tPPtr &) const throw(Veto,Exception);

   double negativeBinomial(int N, double mean, double ek);
   double meanMultiplicity(Energy E);
   int multiplicity(Energy E);

   static ClassDescription<UA5Handler> initUA5Handler;

   /**
    * This is never defined and since it can never be called it isn't 
    * needed. The prototype is defined so the compiler doesn't use the 
    * default = operator.
    */
   UA5Handler& operator=(const UA5Handler &);

   GlobParamPtr        _globalParams;
   ClusterFissionerPtr _clusterFissioner;
   ClusterDecayerPtr   _clusterDecayer;
   PartonSplitterPtr   _split;

   /**
    * The parameters to the mean multiplicity distribution at mass 
    * s^(1/2) given by N1 * s^N2 + N3.
    */
   double  _N1, _N2, _N3;  

   /**
    * These are used to specify the k of the inverse binomial 
    * distribution given by
    *   P(n) = G(n+k)/(n! G(k)) (n_/k)^n/(1+n_/k)^(n+k)
    * where n_ is the mean of the distribution and G(x) is the 
    * Gamma function.
    */
   double  _K1, _K2; 

   /**
    * These are parameters used in the mass distribution of the 
    * clusters given as  (M-m1-m2-_M1) exp(-_M2 * M)  where M is 
    * the decaying cluster mass and m1, m2 are the consituent masses.
    */
   double _M1, _M2;  

   /**
    * This is used to generate pt of a soft cluster. These parameters 
    * control the slope of the distribution of (d,u) clusters, 
    * (s,c) clusters and diquark clusters respectively. 
    * The distribution is given as
    *         P(pt) ~ pt*exp(-b (pt^2+M^2)^(1/2)) 
    * where b is the given by the relevant parameter.
    */
   double _P1, _P2, _P3;  

   /**
    * This is the probability of having a soft underlying event.
    */
   double _probSoft;

   /**
    * This is a parameter used to enhance the CM energy used to 
    * generate the multiplicity distribution.
    */
   double _enhanceCM; 

};

}

namespace ThePEG {

/** This template specialization informs ThePEG about the
 *  base classes of UA5Handler. */
template<>
struct BaseClassTrait<Herwig::UA5Handler,1> { 
  /** Typedef of the first base class of UA5Handler. */
  typedef MultipleInteractionHandler NthBase;
};

/** This template specialization informs ThePEG about the name of
 *  the UA5Handler class and the shared object where it is defined. */
template<>
struct ClassTraits<Herwig::UA5Handler> :
  public ClassTraitsBase<Herwig::UA5Handler> {
  /** Return a platform-independent class name */
    static string className() { return "Herwig++::UA5Handler"; }
  /** Return the name of the shared library be loaded to get
   *  access to the WeakPartonicDecayer class and every other class it uses
   *  (except the base class). */
    static string library() { return "HwUA5.so"; }
};

}

#endif
